import shutil
from typing import Dict, Text

import pytest
from _pytest.monkeypatch import MonkeyPatch
from rasax.community import utils
from rasax.community.services.git_service import GitService
from sanic.testing import SanicTestClient
from requests_async.models import Response
from sqlalchemy.orm import Session

BASE_URL = "/api/projects/default/git"
REPOSITORIES_URL = f"{BASE_URL}/repositories"

TEST_REPOSITORY = {
    "name": "my-repo",
    "repository_url": "my-url",
    "ssh_key": "safe-key",
    "git_service": "Github",
    "git_service_access_token": "token",
    "target_branch": "master",
}


@pytest.fixture()
def test_repository(session: Session) -> Dict:
    from tests.unit.services.test_git_service import RASA_REMOTE_REPOSITORY

    test_repo = TEST_REPOSITORY.copy()
    test_repo.update(RASA_REMOTE_REPOSITORY)
    return test_repo


@pytest.fixture()
def git_service(session: Session):
    git_service = GitService(session)

    yield git_service
    repository_path = git_service.repository_path()
    if repository_path.exists():
        shutil.rmtree(git_service.repository_path())


def test_add_repository(
    test_client: SanicTestClient, auth_header: Dict[Text, Text], test_repository: Dict
) -> None:
    # create repository
    _, response = test_client.post(
        REPOSITORIES_URL, json=test_repository, headers=auth_header
    )  # type: Response

    assert response.status_code == 201

    # Check response body
    response_body = response.json
    assert response_body.pop("id")

    fields_to_check = test_repository.keys() - {"ssh_key", "git_service_access_token"}
    assert all(
        [response_body[field] == test_repository[field] for field in fields_to_check]
    )


def test_add_repository_with_invalid_body(
    test_client: SanicTestClient, auth_header: Dict[Text, Text], test_repository: Dict
) -> None:
    invalid_repository = {"name": "test"}

    _, response = test_client.post(
        REPOSITORIES_URL, json=invalid_repository, headers=auth_header
    )  # type: Response

    assert response.status_code == 400


def test_list_repositories(
    test_client: SanicTestClient, auth_header: Dict[Text, Text], test_repository: Dict
) -> None:
    # Create repository
    _, response = test_client.post(
        REPOSITORIES_URL, json=test_repository, headers=auth_header
    )  # type: Response

    # List repositories
    _, response = test_client.get(
        REPOSITORIES_URL, headers=auth_header
    )  # type: Response

    assert response.status_code == 200
    assert len(response.json) >= 1  # there might be others from previous tests


def test_get_single_repository(
    test_client: SanicTestClient, auth_header: Dict[Text, Text], test_repository: Dict
) -> None:
    # Create repository
    _, response = test_client.post(
        REPOSITORIES_URL, json=test_repository, headers=auth_header
    )  # type: Response
    id_of_created = response.json["id"]

    # Get repository
    _, response = test_client.get(
        f"{REPOSITORIES_URL}/{id_of_created}", headers=auth_header
    )  # type: Response
    response_body = response.json

    fields_to_check = test_repository.keys() - {"ssh_key", "git_service_access_token"}
    assert all(
        [response_body[field] == test_repository[field] for field in fields_to_check]
    )


def test_update_repository(
    test_client: SanicTestClient, auth_header: Dict[Text, Text], test_repository: Dict
) -> None:
    # Create repository
    _, response = test_client.post(
        REPOSITORIES_URL, json=test_repository, headers=auth_header
    )  # type: Response
    id_of_created = response.json["id"]

    updated = test_repository.copy()
    updated_name = "updated 🎉"
    updated["name"] = updated_name

    # Update repository
    _, response = test_client.put(
        f"{REPOSITORIES_URL}/{id_of_created}", json=updated, headers=auth_header
    )  # type: Response

    assert response.status_code == 200

    # Get repository
    _, response = test_client.get(
        f"{REPOSITORIES_URL}/{id_of_created}", headers=auth_header
    )  # type: Response
    response_body = response.json
    assert response_body["name"] == updated_name


def test_update_repository_with_invalid_id(
    test_client: SanicTestClient, auth_header: Dict[Text, Text], test_repository: Dict
) -> None:
    # Update repository
    _, response = test_client.put(
        f"{REPOSITORIES_URL}/56789", json=test_repository, headers=auth_header
    )  # type: Response

    assert response.status_code == 404


def test_delete_repository(
    test_client: SanicTestClient, auth_header: Dict[Text, Text], test_repository: Dict
) -> None:
    # Create repository
    _, response = test_client.post(
        REPOSITORIES_URL, json=test_repository, headers=auth_header
    )  # type: Response
    id_of_created = response.json["id"]

    # Delete repository
    _, response = test_client.delete(
        f"{REPOSITORIES_URL}/{id_of_created}", headers=auth_header
    )  # type: Response

    assert response.status_code == 200

    # Check repository is gone
    _, response = test_client.get(
        f"{REPOSITORIES_URL}/{id_of_created}", headers=auth_header
    )  #

    assert response.status_code == 404


def test_delete_repository_with_invalid_id(
    test_client: SanicTestClient, auth_header: Dict[Text, Text]
) -> None:
    # Delete repository
    _, response = test_client.delete(
        f"{REPOSITORIES_URL}/456789", headers=auth_header
    )  # type: Response

    assert response.status_code == 404


def test_commit_with_target_branch(
    test_client: SanicTestClient, auth_header: Dict[Text, Text]
) -> None:
    pass


def test_commit_with_other_branch(
    test_client: SanicTestClient, auth_header: Dict[Text, Text]
) -> None:
    pass


def test_get_git_status(
    test_client: SanicTestClient,
    auth_header: Dict[Text, Text],
    test_repository: Dict,
    session: Session,
    git_service: GitService,
) -> None:
    # Create repository
    _, response = test_client.post(
        REPOSITORIES_URL, json=test_repository, headers=auth_header
    )  # type: Response

    # No background job is running in the unit tests, so we have to clone ourselves
    git_service.set_repository(response.json["id"])
    git_service.clone()

    _, response = test_client.get(
        f"{REPOSITORIES_URL}/{response.json['id']}/status", headers=auth_header
    )  # type: Response

    assert response.status_code == 200

    response_body = response.json
    assert response_body.pop("time_of_last_pull") > 0
    assert response_body == {
        "is_committing_to_target_branch_allowed": False,
        "is_remote_ahead": False,
        "local_changes": False,
        "current_branch": "master",
    }


def test_change_branch(
    test_client: SanicTestClient,
    auth_header: Dict[Text, Text],
    test_repository: Dict,
    git_service: GitService,
) -> None:

    # Create repository
    _, response = test_client.post(
        REPOSITORIES_URL, json=test_repository, headers=auth_header
    )  # type: Response
    repository_id = response.json["id"]

    # No background job is running in the unit tests, so we have to clone ourselves
    git_service.set_repository(repository_id)
    git_service.clone()

    # Checkout other branch
    other_branch = "so-different"
    git_service.create_branch(other_branch)
    _, response = test_client.put(
        f"{REPOSITORIES_URL}/{repository_id}/branches/{other_branch}",
        headers=auth_header,
    )  # type: Response

    assert response.status_code == 200

    # Get repository status
    _, response = test_client.get(
        f"{REPOSITORIES_URL}/{repository_id}/status", headers=auth_header
    )

    assert response.json["current_branch"] == other_branch


def test_create_commit(
    test_client: SanicTestClient,
    auth_header: Dict[Text, Text],
    test_repository: Dict,
    git_service: GitService,
    monkeypatch: MonkeyPatch,
) -> None:
    # Create repository
    _, response = test_client.post(
        REPOSITORIES_URL, json=test_repository, headers=auth_header
    )  # type: Response
    repository_id = response.json["id"]

    # Make it dirty
    git_service.set_repository(repository_id)
    git_service.clone()
    utils.write_text_to_file(git_service.repository_path() / "new-file.txt", "content")
    assert git_service.is_dirty()

    # Avoid actually pushing to master
    monkeypatch.setattr(GitService, "push", lambda _, branch_name: True)

    # Commit changes (stage, commit, push)
    _, response = test_client.post(
        f"{REPOSITORIES_URL}/{repository_id}/branches/master/commit",
        json=test_repository,
        headers=auth_header,
    )  # type: Response

    assert response.status_code == 201
    assert not git_service.is_dirty()
